import fs from "fs"
import chalk from "chalk"
import gc from "expose-gc/function"

import { connect } from "../src/room.js"

import MockConnection, {
  getUserID,
  getPeerHandle,
  getPeerFootprint,
  send,
  sendListener,
  broadcast,
  broadcastListener,
  terminatePeer,
  destructor,
  addEventListener,
  getEventListener,
} from "../src/connection/MockConnection.js"

import {
  createMessageReceivedEvent,
  handshake,
  syncStep1,
  syncDone,
  dotDraw,
  dotErase,
  pathDraw,
  pathErase,
} from "./benchmark.data.js"

// Adapted from https://github.com/jprichardson/buffer-json (MIT license)

function stringify(value, space) {
  return JSON.stringify(value, replacer, space)
}

function parse(text) {
  return JSON.parse(text, reviver)
}

function replacer(key, value) {
  if (value instanceof Uint8Array) {
    return "base64:" + Buffer.from(value).toString("base64")
  }
  return value
}

function reviver(key, value) {
  if (typeof value == "string" && value.startsWith("base64:")) {
    return Uint8Array.from(Buffer.from(value.slice("base64:".length), "base64"))
  }
  return value
}

function dumpBSON(filename, data) {
  return fs.writeFileSync(filename, stringify(data))
}

function loadBSON(filename) {
  return parse(fs.readFileSync(filename))
}

function printBenchmark(title, iterations, results) {
  process.stdout.write(`\n  ${title} (${iterations} iterations):\n`)

  for (const title in results) {
    const {
      timeLoc,
      encodeRAM,
      packets,
      size,
      timeRem,
      decodeRAM,
      events,
    } = results[title]
    const synchronisation = title == "synchronisation"

    process.stdout.write(
      chalk`    {yellow ⧗} {dim ${title}:} {yellow.inverse ${(
        timeLoc /
        (1e6 * (synchronisation ? 1 : iterations))
      ).toFixed(3)}ms ${synchronisation ? "total" : "/ it"}} + {red.inverse ${(
        encodeRAM /
        (1024 * 1024)
      ).toFixed(
        3,
      )}MB} => {dim ${packets} packet(s)} => {magenta.inverse ${size}B} => {yellow.inverse ${(
        timeRem /
        (1e6 * (synchronisation ? 1 : iterations))
      ).toFixed(3)}ms ${synchronisation ? "total" : "/ it"}} + {red.inverse ${(
        decodeRAM /
        (1024 * 1024)
      ).toFixed(3)}MB} => {dim ${events} event(s)}\n`,
    )
  }

  process.stdout.write(`\n`)
}

function writeBenchmarkHeader(filename, title, results) {
  const columns = ["iterations"]

  for (const title of results) {
    columns.push(`${title}_timeLoc`)
    columns.push(`${title}_encodeRAM`)
    columns.push(`${title}_packets`)
    columns.push(`${title}_size`)
    columns.push(`${title}_timeRem`)
    columns.push(`${title}_decodeRAM`)
    columns.push(`${title}_events`)
  }

  return fs.writeFileSync(
    filename,
    `# Benchmark: ${title}\n# ${columns.join("\t")}\n`,
  )
}

function appendBenchmark(filename, iterations, results) {
  const columns = [iterations]

  for (const title in results) {
    const {
      timeLoc,
      encodeRAM,
      packets,
      size,
      timeRem,
      decodeRAM,
      events,
    } = results[title]

    columns.push(timeLoc)
    columns.push(encodeRAM)
    columns.push(packets)
    columns.push(size)
    columns.push(timeRem)
    columns.push(decodeRAM)
    columns.push(events)
  }

  return fs.appendFileSync(filename, `${columns.join("\t")}\n`)
}

function captureHeapUsage() {
  for (let i = 0; i < 10; i++) {
    gc()
  }

  return process.memoryUsage().heapUsed
}

function runBidirectionalBenchmark(
  BENCHMARK,
  FILENAME,
  ITERATIONSLIST,
  BLOCKSIZE,

  addData,
  eraseData,

  addOnInitFrontend,
  addBroadcastGroupTimeout,
  addOnBroadcastGroup,
  addPacketsFilename,

  eraseOnInitFrontend,
  eraseBroadcastGroupTimeout,
  eraseOnBroadcastGroup,
  erasePacketsFilename,

  syncSendGroupTimeout,
  syncOnSendGroup,
  syncPacketsFilename,

  addOnInitBackend,
  addEventGroupTimeout,
  addOnEventGroup,
  addEventsCache,
  addEventsFilename,

  eraseOnInitBackend,
  eraseEventGroupTimeout,
  eraseOnEventGroup,
  eraseEventsCache,
  eraseEventsFilename,

  syncEventGroupTimeout,
  syncOnEventGroup,
  syncEventsCache,
  syncEventsFilename,
) {
  if (BENCHMARK && FILENAME) {
    writeBenchmarkHeader(FILENAME, BENCHMARK, [
      "addPath",
      "extendErasureIntervals",
      "synchronisation",
    ])
  }

  return ITERATIONSLIST.reduce(
    (promise, ITERATIONS) =>
      promise.then(() => {
        const pathIDs = []
        let prevTime
        let currTime

        let connectRAM // eslint-disable-line no-unused-vars
        let addLocTime = 0
        let addPackets = []
        let addSize = 0
        let addRAM
        let eraseLocTime = 0
        let erasePackets = []
        let eraseSize = 0
        let eraseRAM
        let syncLocTime
        let syncPackets = []
        let syncSize = 0
        let syncRAM
        let disconnectRAM // eslint-disable-line no-unused-vars

        let connectUpdRAM // eslint-disable-line no-unused-vars
        let addRemTime = 0
        let addEvents = []
        let addRemRAM
        let eraseRemTime = 0
        let eraseEvents = []
        let eraseRemRAM
        let disconnectUpdRAM // eslint-disable-line no-unused-vars

        let connectSyncRAM // eslint-disable-line no-unused-vars
        let syncRemTime = 0
        let syncEvents = []
        let syncRemRAM
        let disconnectSyncRAM // eslint-disable-line no-unused-vars

        let timeout

        let addEventListener = null
        let eraseEventListener = null

        let room = null
        let updateRoom = null
        let syncRoom = null

        return (
          // eslint-disable-next-line no-async-promise-executor
          new Promise(async (resolve) => {
            room = await connect("room", MockConnection)
            getEventListener(
              "room",
              "messageReceived",
            )(createMessageReceivedEvent(handshake, "tw-ml"))

            connectRAM = captureHeapUsage()

            return resolve()
          })
            .then(
              () =>
                new Promise((resolve) => {
                  let broadcasts = 0

                  broadcastListener.callback = (channel, message) => {
                    currTime = process.hrtime()

                    expect(channel).toEqual("y-js")
                    expect(message.message instanceof Uint8Array).toBe(true)

                    addPackets[addPackets.length - 1].push(message)
                    addSize += message.message.length

                    clearTimeout(timeout)
                    timeout = setTimeout(() => {
                      broadcasts += 1

                      addLocTime +=
                        (currTime[0] - prevTime[0]) * 1e9 +
                        (currTime[1] - prevTime[1])

                      prevTime = process.hrtime()

                      addOnBroadcastGroup(
                        room,
                        addPackets,
                        pathIDs,
                        addData,
                        ITERATIONS,
                        broadcasts,
                        resolve,
                      )
                    }, addBroadcastGroupTimeout)
                  }

                  prevTime = process.hrtime()

                  addOnInitFrontend(
                    room,
                    addPackets,
                    pathIDs,
                    addData,
                    ITERATIONS,
                    BLOCKSIZE,
                  )
                }),
            )
            .then(() => {
              broadcastListener.callback = null

              dumpBSON(addPacketsFilename, addPackets)
              addPackets = null

              addRAM = captureHeapUsage()
            })
            .then(
              () =>
                new Promise((resolve) => {
                  let broadcasts = 0

                  broadcastListener.callback = (channel, message) => {
                    currTime = process.hrtime()

                    expect(channel).toEqual("y-js")
                    expect(message.message instanceof Uint8Array).toBe(true)

                    erasePackets[erasePackets.length - 1].push(message)
                    eraseSize += message.message.length

                    clearTimeout(timeout)
                    timeout = setTimeout(() => {
                      broadcasts += 1

                      eraseLocTime +=
                        (currTime[0] - prevTime[0]) * 1e9 +
                        (currTime[1] - prevTime[1])

                      prevTime = process.hrtime()

                      eraseOnBroadcastGroup(
                        room,
                        erasePackets,
                        pathIDs,
                        eraseData,
                        ITERATIONS,
                        broadcasts,
                        resolve,
                      )
                    }, eraseBroadcastGroupTimeout)
                  }

                  prevTime = process.hrtime()

                  eraseOnInitFrontend(
                    room,
                    erasePackets,
                    pathIDs,
                    eraseData,
                    ITERATIONS,
                    BLOCKSIZE,
                  )
                }),
            )
            .then(() => {
              broadcastListener.callback = null

              dumpBSON(erasePacketsFilename, erasePackets)
              erasePackets = null

              eraseRAM = captureHeapUsage()
            })
            .then(
              () =>
                new Promise((resolve) => {
                  sendListener.callback = (uid, channel, message) => {
                    const currTime = process.hrtime()

                    expect(uid).toEqual("moritz")
                    expect(channel).toEqual("y-js")
                    expect(message.message instanceof Uint8Array).toBe(true)

                    syncLocTime =
                      (currTime[0] - prevTime[0]) * 1e9 +
                      (currTime[1] - prevTime[1])

                    syncPackets.push(message)
                    syncSize += message.message.length

                    clearTimeout(timeout)
                    timeout = setTimeout(
                      () => syncOnSendGroup(syncPackets, resolve),
                      syncSendGroupTimeout,
                    )
                  }

                  prevTime = process.hrtime()

                  getEventListener(
                    "room",
                    "messageReceived",
                  )(createMessageReceivedEvent(syncStep1))
                }),
            )
            .then(() => {
              sendListener.callback = null

              dumpBSON(syncPacketsFilename, syncPackets)
              syncPackets = null

              syncRAM = captureHeapUsage()

              room.disconnect()
              room = null

              disconnectRAM = captureHeapUsage()
            })
            .then(
              () =>
                // eslint-disable-next-line no-async-promise-executor
                new Promise(async (resolve) => {
                  updateRoom = await connect("update", MockConnection)
                  getEventListener(
                    "update",
                    "messageReceived",
                  )(createMessageReceivedEvent(handshake, "tw-ml"))

                  connectUpdRAM = captureHeapUsage()

                  addPackets = loadBSON(addPacketsFilename)

                  return resolve()
                }),
            )
            .then(
              () =>
                new Promise((resolve) => {
                  let broadcasts = 0

                  let currTime

                  const timeoutCallback = () => {
                    broadcasts += 1

                    addRemTime +=
                      (currTime[0] - prevTime[0]) * 1e9 +
                      (currTime[1] - prevTime[1])

                    prevTime = process.hrtime()

                    addOnEventGroup(
                      addPackets,
                      ITERATIONS,
                      broadcasts,
                      resolve,
                      addEvents,
                      pathIDs,
                      addData,
                    )
                  }

                  addEventListener = (event) => {
                    currTime = process.hrtime()

                    addEvents.push(addEventsCache ? { add: event } : null)

                    clearTimeout(timeout)
                    timeout = setTimeout(timeoutCallback, addEventGroupTimeout)
                  }
                  eraseEventListener = (event) => {
                    currTime = process.hrtime()

                    addEvents.push(addEventsCache ? { erase: event } : null)

                    clearTimeout(timeout)
                    timeout = setTimeout(timeoutCallback, addEventGroupTimeout)
                  }

                  updateRoom.addEventListener(
                    "addOrUpdatePath",
                    addEventListener,
                  )
                  updateRoom.addEventListener(
                    "removedIntervalsChange",
                    eraseEventListener,
                  )

                  prevTime = process.hrtime()

                  addOnInitBackend(addPackets, BLOCKSIZE)
                }),
            )
            .then(() => {
              updateRoom.removeEventListener(
                "addOrUpdatePath",
                addEventListener,
              )
              addEventListener = null

              updateRoom.removeEventListener(
                "removedIntervalsChange",
                eraseEventListener,
              )
              eraseEventListener = null

              addPackets = null

              dumpBSON(addEventsFilename, addEvents)
              addEvents = null

              addRemRAM = captureHeapUsage()

              erasePackets = loadBSON(erasePacketsFilename)
            })
            .then(
              () =>
                new Promise((resolve) => {
                  let broadcasts = 0

                  let currTime

                  const timeoutCallback = () => {
                    broadcasts += 1

                    eraseRemTime +=
                      (currTime[0] - prevTime[0]) * 1e9 +
                      (currTime[1] - prevTime[1])

                    prevTime = process.hrtime()

                    eraseOnEventGroup(
                      erasePackets,
                      ITERATIONS,
                      broadcasts,
                      resolve,
                      eraseEvents,
                      pathIDs,
                      addData,
                      eraseData,
                    )
                  }

                  eraseEventListener = (event) => {
                    currTime = process.hrtime()

                    eraseEvents.push(eraseEventsCache ? { erase: event } : null)

                    clearTimeout(timeout)
                    timeout = setTimeout(
                      timeoutCallback,
                      eraseEventGroupTimeout,
                    )
                  }

                  updateRoom.addEventListener(
                    "removedIntervalsChange",
                    eraseEventListener,
                  )

                  prevTime = process.hrtime()

                  eraseOnInitBackend(erasePackets, BLOCKSIZE)
                }),
            )
            .then(() => {
              updateRoom.removeEventListener(
                "removedIntervalsChange",
                eraseEventListener,
              )
              eraseEventListener = null

              erasePackets = null

              dumpBSON(eraseEventsFilename, eraseEvents)
              eraseEvents = null

              eraseRemRAM = captureHeapUsage()

              updateRoom.disconnect()
              updateRoom = null

              disconnectUpdRAM = captureHeapUsage()
            })
            .then(
              () =>
                // eslint-disable-next-line no-async-promise-executor
                new Promise(async (resolve) => {
                  syncRoom = await connect("sync", MockConnection)
                  getEventListener(
                    "sync",
                    "messageReceived",
                  )(createMessageReceivedEvent(handshake, "tw-ml"))

                  connectSyncRAM = captureHeapUsage()

                  syncPackets = loadBSON(syncPacketsFilename)

                  return resolve()
                }),
            )
            .then(
              () =>
                new Promise((resolve) => {
                  addEventListener = (event) => {
                    const currTime = process.hrtime()

                    syncRemTime =
                      (currTime[0] - prevTime[0]) * 1e9 +
                      (currTime[1] - prevTime[1])

                    syncEvents.push(syncEventsCache ? { add: event } : null)

                    clearTimeout(timeout)
                    timeout = setTimeout(
                      () =>
                        syncOnEventGroup(
                          resolve,
                          syncEvents,
                          pathIDs,
                          addData,
                          eraseData,
                        ),
                      syncEventGroupTimeout,
                    )
                  }
                  eraseEventListener = (event) => {
                    const currTime = process.hrtime()

                    syncRemTime =
                      (currTime[0] - prevTime[0]) * 1e9 +
                      (currTime[1] - prevTime[1])

                    syncEvents.push(syncEventsCache ? { erase: event } : null)

                    clearTimeout(timeout)
                    timeout = setTimeout(
                      () =>
                        syncOnEventGroup(
                          resolve,
                          syncEvents,
                          pathIDs,
                          addData,
                          eraseData,
                        ),
                      syncEventGroupTimeout,
                    )
                  }

                  syncRoom.addEventListener("addOrUpdatePath", addEventListener)
                  syncRoom.addEventListener(
                    "removedIntervalsChange",
                    eraseEventListener,
                  )

                  prevTime = process.hrtime()

                  for (const syncPacket of syncPackets) {
                    getEventListener(
                      "sync",
                      "messageReceived",
                    )(createMessageReceivedEvent(syncPacket))
                  }
                }),
            )
            .then(() => {
              syncRoom.removeEventListener("addOrUpdatePath", addEventListener)
              addEventListener = null

              syncRoom.removeEventListener(
                "removedIntervalsChange",
                eraseEventListener,
              )
              eraseEventListener = null

              syncPackets = null

              dumpBSON(syncEventsFilename, syncEvents)
              syncEvents = null

              syncRemRAM = captureHeapUsage()

              syncRoom.disconnect()
              syncRoom = null

              disconnectSyncRAM = captureHeapUsage()
            })
            .then(() => {
              if (!BENCHMARK) {
                return
              }

              addPackets = loadBSON(addPacketsFilename).reduce(
                (sum, packets) => sum + packets.length,
                0,
              )
              erasePackets = loadBSON(erasePacketsFilename).reduce(
                (sum, packets) => sum + packets.length,
                0,
              )
              syncPackets = loadBSON(syncPacketsFilename).length

              addEvents = loadBSON(addEventsFilename).length
              eraseEvents = loadBSON(eraseEventsFilename).length
              syncEvents = loadBSON(syncEventsFilename).length

              const results = {
                addPath: {
                  timeLoc: addLocTime,
                  encodeRAM: addRAM,
                  packets: addPackets,
                  size: addSize,
                  timeRem: addRemTime,
                  decodeRAM: addRemRAM,
                  events: addEvents,
                },
                extendErasureIntervals: {
                  timeLoc: eraseLocTime,
                  encodeRAM: eraseRAM,
                  packets: erasePackets,
                  size: eraseSize,
                  timeRem: eraseRemTime,
                  decodeRAM: eraseRemRAM,
                  events: eraseEvents,
                },
                synchronisation: {
                  timeLoc: syncLocTime,
                  encodeRAM: syncRAM,
                  packets: syncPackets,
                  size: syncSize,
                  timeRem: syncRemTime,
                  decodeRAM: syncRemRAM,
                  events: syncEvents,
                },
              }

              printBenchmark(BENCHMARK, ITERATIONS, results)
              appendBenchmark(FILENAME, ITERATIONS, results)
            })
        )
      }),
    Promise.resolve(),
  )
}

function addOnInitFrontendSequential(
  room,
  addPackets,
  pathIDs,
  addData,
  ITERATIONS, // eslint-disable-line no-unused-vars
  BLOCKSIZE, // eslint-disable-line no-unused-vars
) {
  addPackets.push([])

  const drawPathID = room.addPath(addData[0])

  pathIDs.push(drawPathID)
  for (let i = 1; i < addData.length; i++) {
    room.extendPath(drawPathID, addData[i])
  }
}

function addOnBroadcastGroupSequential(
  room,
  addPackets,
  pathIDs,
  addData,
  ITERATIONS,
  broadcasts,
  resolve,
) {
  if (broadcasts < ITERATIONS) {
    addPackets.push([])

    const drawPathID = room.addPath(addData[0])
    pathIDs.push(drawPathID)

    for (let i = 1; i < addData.length; i++) {
      room.extendPath(drawPathID, addData[i])
    }
  } else {
    resolve()
  }
}

function eraseOnInitFrontendSequential(
  room,
  erasePackets,
  pathIDs,
  eraseData,
  ITERATIONS, // eslint-disable-line no-unused-vars
  BLOCKSIZE, // eslint-disable-line no-unused-vars
) {
  erasePackets.push([])

  const erasePathID = pathIDs[0]

  for (let i = 0; i < eraseData.length; i++) {
    room.extendErasureIntervals(erasePathID, eraseData[i][0], eraseData[i][1])
  }
}

function eraseOnBroadcastGroupSequential(
  room,
  erasePackets,
  pathIDs,
  eraseData,
  ITERATIONS,
  broadcasts,
  resolve,
) {
  if (broadcasts < ITERATIONS) {
    erasePackets.push([])

    const erasePathID = pathIDs[broadcasts]

    for (let i = 0; i < eraseData.length; i++) {
      room.extendErasureIntervals(erasePathID, eraseData[i][0], eraseData[i][1])
    }
  } else {
    resolve()
  }
}

function syncOnSendGroup(syncPackets, resolve) {
  resolve()
}

function addOnInitBackendSequential(
  addPackets,
  BLOCKSIZE, // eslint-disable-line no-unused-vars
) {
  for (const packet of addPackets[0]) {
    getEventListener(
      "update",
      "messageReceived",
    )(createMessageReceivedEvent(packet))
  }
}

function addOnEventGroupSequential(
  addPackets,
  ITERATIONS,
  broadcasts,
  resolve,
  addEvents, // eslint-disable-line no-unused-vars
  pathIDs, // eslint-disable-line no-unused-vars
  addData, // eslint-disable-line no-unused-vars
) {
  if (broadcasts >= ITERATIONS) {
    return resolve()
  }

  for (const packet of addPackets[broadcasts]) {
    getEventListener(
      "update",
      "messageReceived",
    )(createMessageReceivedEvent(packet))
  }
}

function eraseOnInitBackendSequential(
  erasePackets,
  BLOCKSIZE, // eslint-disable-line no-unused-vars
) {
  for (const packet of erasePackets[0]) {
    getEventListener(
      "update",
      "messageReceived",
    )(createMessageReceivedEvent(packet))
  }
}

function eraseOnEventGroupSequential(
  erasePackets,
  ITERATIONS,
  broadcasts,
  resolve,
  eraseEvents, // eslint-disable-line no-unused-vars
  pathIDs, // eslint-disable-line no-unused-vars
  addData, // eslint-disable-line no-unused-vars
  eraseData, // eslint-disable-line no-unused-vars
) {
  if (broadcasts >= ITERATIONS) {
    return resolve()
  }

  for (const packet of erasePackets[broadcasts]) {
    getEventListener(
      "update",
      "messageReceived",
    )(createMessageReceivedEvent(packet))
  }
}

function syncOnEventGroup(
  resolve,
  syncEvents, // eslint-disable-line no-unused-vars
  pathIDs, // eslint-disable-line no-unused-vars
  addData, // eslint-disable-line no-unused-vars
  eraseData, // eslint-disable-line no-unused-vars
) {
  resolve()
}

function addOnInitFrontendParallel(
  room,
  addPackets,
  pathIDs,
  addData,
  ITERATIONS,
  BLOCKSIZE,
) {
  addPackets.push([])

  // Necessary to allow yjs to execute transactions (majority of processing time)
  function addPath(sj) {
    if (sj >= ITERATIONS) return

    for (let j = sj; j < Math.min(sj + BLOCKSIZE, ITERATIONS); j++) {
      const drawPathID = room.addPath(addData[0])

      pathIDs.push(drawPathID)
      for (let i = 1; i < addData.length; i++) {
        room.extendPath(drawPathID, addData[i])
      }
    }

    setTimeout(addPath, 0, sj + BLOCKSIZE)
  }

  addPath(0)
}

function addOnBroadcastGroupParallel(
  room,
  addPackets,
  pathIDs,
  addData,
  ITERATIONS,
  broadcasts,
  resolve,
) {
  resolve()
}

function eraseOnInitFrontendParallel(
  room,
  erasePackets,
  pathIDs,
  eraseData,
  ITERATIONS,
  BLOCKSIZE,
) {
  erasePackets.push([])

  // Necessary to allow yjs to execute transactions (majority of processing time)
  function erasePath(sj) {
    if (sj >= ITERATIONS) return

    for (let j = sj; j < Math.min(sj + BLOCKSIZE, ITERATIONS); j++) {
      const erasePathID = pathIDs[j]

      for (let i = 0; i < eraseData.length; i++) {
        room.extendErasureIntervals(
          erasePathID,
          eraseData[i][0],
          eraseData[i][1],
        )
      }
    }

    setTimeout(erasePath, 0, sj + BLOCKSIZE)
  }

  erasePath(0)
}

function eraseOnBroadcastGroupParallel(
  room,
  erasePackets,
  pathIDs,
  eraseData,
  ITERATIONS,
  broadcasts,
  resolve,
) {
  resolve()
}

function addOnInitBackendParallel(addPackets, BLOCKSIZE) {
  const packets = addPackets[0]

  function addPath(sj) {
    if (sj >= packets.length) return

    for (let j = sj; j < Math.min(sj + BLOCKSIZE, packets.length); j++) {
      getEventListener(
        "update",
        "messageReceived",
      )(createMessageReceivedEvent(packets[j]))
    }

    setTimeout(addPath, 0, sj + BLOCKSIZE)
  }

  addPath(0)
}

function addOnEventGroupParallel(
  addPackets,
  ITERATIONS,
  broadcasts,
  resolve,
  addEvents, // eslint-disable-line no-unused-vars
  pathIDs, // eslint-disable-line no-unused-vars
  addData, // eslint-disable-line no-unused-vars
) {
  resolve()
}

function eraseOnInitBackendParallel(erasePackets, BLOCKSIZE) {
  const packets = erasePackets[0]

  function erasePath(sj) {
    if (sj >= packets.length) return

    for (let j = sj; j < Math.min(sj + BLOCKSIZE, packets.length); j++) {
      getEventListener(
        "update",
        "messageReceived",
      )(createMessageReceivedEvent(packets[j]))
    }

    setTimeout(erasePath, 0, sj + BLOCKSIZE)
  }

  erasePath(0)
}

function eraseOnEventGroupParallel(
  erasePackets,
  ITERATIONS,
  broadcasts,
  resolve,
  eraseEvents, // eslint-disable-line no-unused-vars
  pathIDs, // eslint-disable-line no-unused-vars
  addData, // eslint-disable-line no-unused-vars
  eraseData, // eslint-disable-line no-unused-vars
) {
  resolve()
}

function syncOnSendGroupVerify(syncPackets, resolve) {
  let syncDonePacket = -1

  syncPackets.forEach((packet, i) => {
    if (
      packet.message.length == syncDone.message.length &&
      JSON.stringify(Object.assign({}, packet, { uuid: undefined })) ==
        JSON.stringify(syncDone)
    ) {
      expect(syncDonePacket).toEqual(-1)

      syncDonePacket = i
    }
  })

  expect(syncDonePacket).toEqual(syncPackets.length - 1)

  resolve()
}

function addOnEventGroupVerify(
  addPackets,
  ITERATIONS,
  broadcasts,
  resolve,
  addEvents,
  pathIDs,
  addData,
) {
  const updatePaths = {}
  const updateIntervals = {}

  for (const event of addEvents) {
    expect(!event.add + !event.erase).toBe(1)

    if (event.add) {
      const {
        detail: { id, points },
      } = event.add

      updatePaths[id] = points
    } else if (event.erase) {
      const {
        detail: { id, intervals },
      } = event.erase

      updateIntervals[id] = intervals
    }
  }

  expect(updatePaths).toStrictEqual({ [pathIDs[0]]: addData })
  expect(updateIntervals).toStrictEqual({ [pathIDs[0]]: {} })

  resolve()
}

function eraseOnEventGroupVerify(
  erasePackets,
  ITERATIONS,
  broadcasts,
  resolve,
  eraseEvents,
  pathIDs,
  addData,
  eraseData, // eslint-disable-line no-unused-vars
) {
  const updateIntervals = {}

  for (const event of eraseEvents) {
    expect(!event.add).toBe(true)
    expect(!event.erase).toBe(false)

    const {
      detail: { id, intervals },
    } = event.erase

    updateIntervals[id] = intervals
  }

  expect(updateIntervals).toStrictEqual({
    [pathIDs[0]]: Object.assign(
      {},
      new Array(addData.length).fill([[0, 0 + (addData.length > 1)]]),
    ),
  })

  resolve()
}

function syncOnEventGroupVerify(
  resolve,
  syncEvents,
  pathIDs,
  addData,
  eraseData, // eslint-disable-line no-unused-vars
) {
  const syncPaths = {}
  const syncIntervals = {}

  for (const event of syncEvents) {
    expect(!event.add + !event.erase).toBe(1)

    if (event.add) {
      const {
        detail: { id, points },
      } = event.add

      syncPaths[id] = points
    } else if (event.erase) {
      const {
        detail: { id, intervals },
      } = event.erase

      syncIntervals[id] = intervals
    }
  }

  expect(syncPaths).toStrictEqual({ [pathIDs[0]]: addData })
  expect(syncIntervals).toStrictEqual({
    [pathIDs[0]]: Object.assign(
      {},
      new Array(addData.length).fill([[0, 0 + (addData.length > 1)]]),
    ),
  })

  resolve()
}

describe("drawing app mesh", () => {
  beforeEach(() => {
    getUserID.mockClear()
    getPeerHandle.mockClear()
    getPeerFootprint.mockClear()
    send.mockClear()
    broadcast.mockClear()
    terminatePeer.mockClear()
    destructor.mockClear()
    addEventListener.mockClear()
    MockConnection.mockClear()

    captureHeapUsage()
  })

  const ITERATIONSLIST = [10, 25, 50, 75, 100, 250, 500]
  const BLOCKSIZE = 10

  jest.setTimeout(1000 * 60 * 60)

  it("benchmarks a dot draw and erase update sequentially", () => {
    return runBidirectionalBenchmark(
      "dot draw and erase [sequential]" /* BENCHMARK */,
      "plots/dot-seq-benchmark.tsv" /* FILENAME */,
      ITERATIONSLIST /* ITERATIONSLIST */,
      BLOCKSIZE /* BLOCKSIZE */,
      dotDraw /* addData */,
      dotErase /* eraseData */,
      addOnInitFrontendSequential /* addOnInitFrontend */,
      100 /* addBroadcastGroupTimeout */,
      addOnBroadcastGroupSequential /* addOnBroadcastGroup */,
      ".dot-seq-add-packets.json" /* addPacketsFilename */,
      eraseOnInitFrontendSequential /* eraseOnInitFrontend */,
      100 /* eraseBroadcastGroupTimeout */,
      eraseOnBroadcastGroupSequential /* eraseOnBroadcastGroup */,
      ".dot-seq-erase-packets.json" /* erasePacketsFilename */,
      1000 /* syncSendGroupTimeout */,
      syncOnSendGroup /* syncOnSendGroup */,
      ".dot-seq-sync-packets.json" /* syncPacketsFilename */,
      addOnInitBackendSequential /* addOnInitBackend */,
      100 /* addEventGroupTimeout */,
      addOnEventGroupSequential /* addOnEventGroup */,
      false /* addEventsCache */,
      ".dot-seq-add-events.json" /* addEventsFilename */,
      eraseOnInitBackendSequential /* eraseOnInitBackend */,
      100 /* eraseEventGroupTimeout */,
      eraseOnEventGroupSequential /* eraseOnEventGroupTimeout */,
      false /* eraseEventsCache */,
      ".dot-seq-erase-events.json" /* eraseEventsFilename */,
      1000 /* syncEventGroupTimeout */,
      syncOnEventGroup /* syncOnEventGroup */,
      false /* syncEventsCache */,
      ".dot-seq-sync-events.json" /* syncEventsFilename */,
    )
  })

  it("benchmarks a dot draw and erase update in parallel", () => {
    return runBidirectionalBenchmark(
      "dot draw and erase [parallel]" /* BENCHMARK */,
      "plots/dot-par-benchmark.tsv" /* FILENAME */,
      ITERATIONSLIST /* ITERATIONSLIST */,
      BLOCKSIZE /* BLOCKSIZE */,
      dotDraw /* addData */,
      dotErase /* eraseData */,
      addOnInitFrontendParallel /* addOnInitFrontend */,
      1000 /* addBroadcastGroupTimeout */,
      addOnBroadcastGroupParallel /* addOnBroadcastGroup */,
      ".dot-par-add-packets.json" /* addPacketsFilename */,
      eraseOnInitFrontendParallel /* eraseOnInitFrontend */,
      1000 /* eraseBroadcastGroupTimeout */,
      eraseOnBroadcastGroupParallel /* eraseOnBroadcastGroup */,
      ".dot-par-erase-packets.json" /* erasePacketsFilename */,
      1000 /* syncSendGroupTimeout */,
      syncOnSendGroup /* syncOnSendGroup */,
      ".dot-par-sync-packets.json" /* syncPacketsFilename */,
      addOnInitBackendParallel /* addOnInitBackend */,
      1000 /* addEventGroupTimeout */,
      addOnEventGroupParallel /* addOnEventGroup */,
      false /* addEventsCache */,
      ".dot-par-add-events.json" /* addEventsFilename */,
      eraseOnInitBackendParallel /* eraseOnInitBackend */,
      1000 /* eraseEventGroupTimeout */,
      eraseOnEventGroupParallel /* eraseOnEventGroupTimeout */,
      false /* eraseEventsCache */,
      ".dot-par-erase-events.json" /* eraseEventsFilename */,
      1000 /* syncEventGroupTimeout */,
      syncOnEventGroup /* syncOnEventGroup */,
      false /* syncEventsCache */,
      ".dot-par-sync-events.json" /* syncEventsFilename */,
    )
  })

  it("benchmarks a path draw and erase update sequentially", () => {
    return runBidirectionalBenchmark(
      "path draw and erase [sequential]" /* BENCHMARK */,
      "plots/path-seq-benchmark.tsv" /* FILENAME */,
      ITERATIONSLIST /* ITERATIONSLIST */,
      BLOCKSIZE /* BLOCKSIZE */,
      pathDraw /* addData */,
      pathErase /* eraseData */,
      addOnInitFrontendSequential /* addOnInitFrontend */,
      100 /* addBroadcastGroupTimeout */,
      addOnBroadcastGroupSequential /* addOnBroadcastGroup */,
      ".path-seq-add-packets.json" /* addPacketsFilename */,
      eraseOnInitFrontendSequential /* eraseOnInitFrontend */,
      100 /* eraseBroadcastGroupTimeout */,
      eraseOnBroadcastGroupSequential /* eraseOnBroadcastGroup */,
      ".path-seq-erase-packets.json" /* erasePacketsFilename */,
      1000 /* syncSendGroupTimeout */,
      syncOnSendGroup /* syncOnSendGroup */,
      ".path-seq-sync-packets.json" /* syncPacketsFilename */,
      addOnInitBackendSequential /* addOnInitBackend */,
      100 /* addEventGroupTimeout */,
      addOnEventGroupSequential /* addOnEventGroup */,
      false /* addEventsCache */,
      ".path-seq-add-events.json" /* addEventsFilename */,
      eraseOnInitBackendSequential /* eraseOnInitBackend */,
      100 /* eraseEventGroupTimeout */,
      eraseOnEventGroupSequential /* eraseOnEventGroupTimeout */,
      false /* eraseEventsCache */,
      ".path-seq-erase-events.json" /* eraseEventsFilename */,
      1000 /* syncEventGroupTimeout */,
      syncOnEventGroup /* syncOnEventGroup */,
      false /* syncEventsCache */,
      ".path-seq-sync-events.json" /* syncEventsFilename */,
    )
  })

  it("benchmarks a path draw and erase update in parallel", () => {
    return runBidirectionalBenchmark(
      "path draw and erase [parallel]" /* BENCHMARK */,
      "plots/path-par-benchmark.tsv" /* FILENAME */,
      ITERATIONSLIST /* ITERATIONSLIST */,
      BLOCKSIZE /* BLOCKSIZE */,
      pathDraw /* addData */,
      pathErase /* eraseData */,
      addOnInitFrontendParallel /* addOnInitFrontend */,
      5000 /* addBroadcastGroupTimeout */,
      addOnBroadcastGroupParallel /* addOnBroadcastGroup */,
      ".path-par-add-packets.json" /* addPacketsFilename */,
      eraseOnInitFrontendParallel /* eraseOnInitFrontend */,
      5000 /* eraseBroadcastGroupTimeout */,
      eraseOnBroadcastGroupParallel /* eraseOnBroadcastGroup */,
      ".path-par-erase-packets.json" /* erasePacketsFilename */,
      5000 /* syncSendGroupTimeout */,
      syncOnSendGroup /* syncOnSendGroup */,
      ".path-par-sync-packets.json" /* syncPacketsFilename */,
      addOnInitBackendParallel /* addOnInitBackend */,
      5000 /* addEventGroupTimeout */,
      addOnEventGroupParallel /* addOnEventGroup */,
      false /* addEventsCache */,
      ".path-par-add-events.json" /* addEventsFilename */,
      eraseOnInitBackendParallel /* eraseOnInitBackend */,
      5000 /* eraseEventGroupTimeout */,
      eraseOnEventGroupParallel /* eraseOnEventGroupTimeout */,
      false /* eraseEventsCache */,
      ".path-par-erase-events.json" /* eraseEventsFilename */,
      5000 /* syncEventGroupTimeout */,
      syncOnEventGroup /* syncOnEventGroup */,
      false /* syncEventsCache */,
      ".path-par-sync-events.json" /* syncEventsFilename */,
    )
  })

  it("communicates a single draw and erase update", () => {
    return runBidirectionalBenchmark(
      null /* BENCHMARK */,
      null /* FILENAME */,
      [1] /* ITERATIONSLIST */,
      BLOCKSIZE /* BLOCKSIZE */,
      dotDraw /* addData */,
      dotErase /* eraseData */,
      addOnInitFrontendSequential /* addOnInitFrontend */,
      1000 /* addBroadcastGroupTimeout */,
      addOnBroadcastGroupParallel /* addOnBroadcastGroup */,
      ".dot-ver-add-packets.json" /* addPacketsFilename */,
      eraseOnInitFrontendSequential /* eraseOnInitFrontend */,
      1000 /* eraseBroadcastGroupTimeout */,
      eraseOnBroadcastGroupParallel /* eraseOnBroadcastGroup */,
      ".dot-ver-erase-packets.json" /* erasePacketsFilename */,
      1000 /* syncSendGroupTimeout */,
      syncOnSendGroupVerify /* syncOnSendGroup */,
      ".dot-ver-sync-packets.json" /* syncPacketsFilename */,
      addOnInitBackendSequential /* addOnInitBackend */,
      1000 /* addEventGroupTimeout */,
      addOnEventGroupVerify /* addOnEventGroup */,
      true /* addEventsCache */,
      ".dot-ver-add-events.json" /* addEventsFilename */,
      eraseOnInitBackendSequential /* eraseOnInitBackend */,
      1000 /* eraseEventGroupTimeout */,
      eraseOnEventGroupVerify /* eraseOnEventGroupTimeout */,
      true /* eraseEventsCache */,
      ".dot-ver-erase-events.json" /* eraseEventsFilename */,
      1000 /* syncEventGroupTimeout */,
      syncOnEventGroupVerify /* syncOnEventGroup */,
      true /* syncEventsCache */,
      ".dot-ver-sync-events.json" /* syncEventsFilename */,
    )
  })

  it("communicates a path draw and erase update", () => {
    return runBidirectionalBenchmark(
      null /* BENCHMARK */,
      null /* FILENAME */,
      [1] /* ITERATIONSLIST */,
      BLOCKSIZE /* BLOCKSIZE */,
      pathDraw /* addData */,
      pathErase /* eraseData */,
      addOnInitFrontendSequential /* addOnInitFrontend */,
      1000 /* addBroadcastGroupTimeout */,
      addOnBroadcastGroupParallel /* addOnBroadcastGroup */,
      ".path-ver-add-packets.json" /* addPacketsFilename */,
      eraseOnInitFrontendSequential /* eraseOnInitFrontend */,
      1000 /* eraseBroadcastGroupTimeout */,
      eraseOnBroadcastGroupParallel /* eraseOnBroadcastGroup */,
      ".path-ver-erase-packets.json" /* erasePacketsFilename */,
      1000 /* syncSendGroupTimeout */,
      syncOnSendGroupVerify /* syncOnSendGroup */,
      ".path-ver-sync-packets.json" /* syncPacketsFilename */,
      addOnInitBackendSequential /* addOnInitBackend */,
      1000 /* addEventGroupTimeout */,
      addOnEventGroupVerify /* addOnEventGroup */,
      true /* addEventsCache */,
      ".path-ver-add-events.json" /* addEventsFilename */,
      eraseOnInitBackendSequential /* eraseOnInitBackend */,
      1000 /* eraseEventGroupTimeout */,
      eraseOnEventGroupVerify /* eraseOnEventGroupTimeout */,
      true /* eraseEventsCache */,
      ".path-ver-erase-events.json" /* eraseEventsFilename */,
      1000 /* syncEventGroupTimeout */,
      syncOnEventGroupVerify /* syncOnEventGroup */,
      true /* syncEventsCache */,
      ".path-ver-sync-events.json" /* syncEventsFilename */,
    )
  })
})
